"""Utils to fill templates at the start of the container.

Template must be filled :
    - after volume mounts and environment initialization,
    - before application start script.
"""
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from nua.lib.actions import jinja2_render_file


def render_config_templates(data: dict[str, Any]):
    templater = Templater(data)
    templater.render_config_templates()


@dataclass
class Templater:
    data: dict[str, Any]

    def render_config_templates(self):
        """Find files and templates in the /nua/templates folder and fill them.

        Destination files may be on Docker-mounted volumes.

        This function is expected to be run as root. Files are stored with
        755 rights. Files ending with .j2 are considered as Jinja2
        templates.
        """
        src_folder = Path("/nua/templates")
        if not src_folder.is_dir():
            return
        self.data.update(os.environ)
        self._iterate_template_files(src_folder)

    def _iterate_template_files(self, src_folder: Path):
        for file in src_folder.rglob("*"):
            if not file.is_file():
                continue
            if file.name.startswith("."):
                continue
            dest = Path("/") / file.relative_to(src_folder)
            if file.name.endswith(".j2"):
                self._render_template(file, dest.with_suffix(""))
            else:
                self._copy_file(file, dest)

    def _copy_file(self, src: Path, dest: Path):
        dest.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
        dest.write_bytes(src.read_bytes())
        dest.chmod(0o644)

    def _render_template(self, template: Path, dest: Path):
        dest.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
        jinja2_render_file(template, dest, self.data)
        dest.chmod(0o644)



"""Apply automated installation detection heuristics.
"""
from importlib import import_module
from importlib import resources as rso
from operator import attrgetter
from pathlib import Path
from typing import Any

from nua.lib.backports import chdir
from nua.lib.panic import info, show
from nua.lib.tool.state import verbosity

from .detectors.base_detector import BaseDetector

DETECTORS_DIRS = ("nua.agent.detectors",)
detector_classes: set[BaseDetector] = set()


def detect_and_install(directory: str | Path | None) -> bool:
    """Apply automated installation detection heuristics."""
    if not detector_classes:
        register_detectors()
    if directory:
        path = Path(directory).resolve()
    else:
        path = Path(".").resolve()
    with verbosity(2):
        info("Detect and install in ", path)
    with chdir(path):
        return auto_install()


def auto_install() -> bool:
    for auto_installer in sorted(detector_classes, key=attrgetter("priority")):
        if auto_installer.detect():
            with verbosity(2):
                msg = auto_installer.info()
                show(f"{msg} detected")
            auto_installer.install_with_build_packages()
            return True
    with verbosity(2):
        show("No automated installation found")
    return False


def register_detectors():
    for dir in DETECTORS_DIRS:
        for file in rso.files(dir).iterdir():
            path = Path(str(file))
            if path.suffix != ".py" or path.stem.startswith("_"):
                continue
            module_path = f"nua.agent.detectors.{path.stem}"
            import_module(module_path)


def register_detector(cls: Any):
    detector_classes.add(cls)



"""Agent libray of Nua, embedded in each Nua Docker image."""

# Version of module available at first level, __version__ actually
# computed from pyproject data in version.py module

# import packaging  # to ensure the module is installed

from .version import __version__  # noqa: F401

# assert packaging  # to avoid flake8 warning



"""Base class for detect and install project."""
import abc

from nua.lib.actions import install_build_packages, installed_packages


class BaseDetector(abc.ABC):
    """Base class for detect and install project.

    BaseDetector.priority: permits to sort the tests order (lower value is higher
    priority) of each detector upon a source directory. Thus testing dual
    installation (ie: python+node) before single ones."""

    message: str = ""
    priority: int = 100
    build_packages: list[str] = []
    run_packages: list[str] = []  # currently not implemented

    @classmethod
    def info(cls) -> str:
        return cls.message

    @classmethod
    def detect(cls) -> bool:
        return False

    @classmethod
    def install_with_build_packages(cls) -> None:
        if cls.build_packages:
            with install_build_packages(
                cls.build_packages,
                installed=installed_packages(),
                keep_lists=True,
            ):
                cls.install()
        else:
            cls.install()

    @classmethod
    def install(cls) -> None:
        pass



"""Auto install Python wheels."""
from pathlib import Path

from nua.lib.actions import pip_install

from ..auto_install import register_detector
from .base_detector import BaseDetector


class PythonWheels(BaseDetector):
    message: str = "Python wheels"
    priority: int = 110  # test order after PythonSource

    @classmethod
    def detect(cls) -> bool:
        root = Path(".").resolve()
        return bool(list(root.glob("*.whl")))

    @classmethod
    def install(cls) -> None:
        pip_install(["*.whl"], user="nua")


register_detector(PythonWheels)



"""Install the configured application inside the container.

- information come from a mandatory local file: "nua-config.toml|json|yaml|yml"
- origin may be a source tar.gz or a git repository, python wheel
- build locally if source is python package
"""
import json
import logging
import os
from copy import deepcopy
from pathlib import Path
from pprint import pformat
from shutil import copy2

from nua.lib.actions import (
    apt_remove_lists,
    install_build_packages,
    install_git_source,
    install_meta_packages,
    install_packages,
    install_pip_packages,
    install_source,
    installed_packages,
)
from nua.lib.backports import chdir
from nua.lib.constants import (
    NUA_APP_PATH,
    NUA_BUILD_PATH,
    NUA_METADATA_PATH,
    NUA_SCRIPTS_PATH,
)
from nua.lib.exec import exec_as_nua, exec_as_root
from nua.lib.nua_config import NuaConfig, force_list, hyphen_get
from nua.lib.panic import Abort, info, show, vprint, warning
from nua.lib.shell import chmod_r, chown_r, mkdir_p, rm_fr, sh
from nua.lib.tool.state import (
    set_packages_updated,
    set_verbosity,
    verbosity,
    verbosity_level,
)

from ..auto_install import detect_and_install

logging.basicConfig(level=logging.INFO)


class AppBuilder:
    """Class to hold config and other state information during build."""

    def __init__(self):
        set_packages_updated(False)

        self.build_dir = Path(NUA_BUILD_PATH)
        if not self.build_dir.is_dir():
            raise Abort(f"Build directory does not exist: '{self.build_dir}'")

        chdir(self.build_dir)
        self.config = NuaConfig(self.build_dir)
        self.source = Path()

    def build(self):
        self.make_dirs()
        with chdir(self.config.root_dir):
            self.pre_build()
            with verbosity(1):
                info("******** Stage: build")
            with install_build_packages(
                self.config.build_packages,
                installed=installed_packages(),
            ):
                code_installed = self.install_project_code()
                chown_r("/nua/build", "nua")
                built = self.run_build_script(code_installed)
                if (code_installed or built) and os.getuid() == 0:
                    chown_r("/nua/build", "nua")

        self.post_build()
        self.test_build()
        with verbosity(1):
            show("******** Build done.")

    def collect_meta_packages(self) -> list[str]:
        """Return meta packages collected from the nua-config requirements
        and connectors required by providers."""
        collected_set = set()
        for provider in self.config.providers:
            collected_set.update(force_list(provider.get("connector") or ""))
        collected_set.update(force_list(self.config.meta_packages))
        return sorted(package for package in collected_set if package)

    def collect_packages(self) -> list[str]:
        """Return packages collected from the nua-config requirements."""
        collected_set = set()
        for provider in self.config.providers:
            collected_set.update(force_list(provider.get("packages", [])))
        collected_set.update(force_list(self.config.packages))
        return sorted(package for package in collected_set if package)

    def pre_build(self):
        """Process installation of packages prior to running install script."""
        with verbosity(1):
            info("******** Stage: pre-build")
        collection = self.collect_meta_packages()
        if collection:
            install_meta_packages(collection, keep_lists=True)
        collection = self.collect_packages()
        if collection:
            with verbosity(1):
                vprint(f"Install packages:  {collection}")
            install_packages(collection, keep_lists=True)
        chown_r("/nua/venv", "nua")

    def make_dirs(self):
        self._make_dir_user_nua(NUA_APP_PATH)
        self._make_dir_user_nua(NUA_METADATA_PATH)
        self.make_custom_document_root()

    def make_custom_document_root(self):
        """If the app defines a specific document root  (i.e.
        /var/www/html)."""
        document_root = hyphen_get(self.config.build, "document-root")
        if not document_root:
            return
        path = Path(document_root)
        rm_fr(path)
        self._make_dir_user_nua(path)

    @staticmethod
    def _make_dir_user_nua(directory: str | Path):
        mkdir_p(directory)
        chown_r(directory, "nua")
        chmod_r(directory, 0o644, 0o755)

    def post_build(self):
        with verbosity(1):
            info("******** Stage: post-build")
        self.merge_files()
        if os.getuid() == 0:
            chown_r("/nua/build", "nua")
        apt_remove_lists()
        self.copy_metadata()
        self.make_start_script()

    def copy_metadata(self):
        """Dump the content of the nua-config file in /nua/metadata/nua-
        config.json."""
        self.config.dump_json(NUA_METADATA_PATH)

    def make_start_script(self):
        script_dir = Path(NUA_SCRIPTS_PATH)
        script_dir.mkdir(mode=0o755, parents=True, exist_ok=True)
        if self.config.start_command:
            with verbosity(2):
                vprint("Writing start script from 'start' content")
            return self._write_start_script(self.config.start_command)
        path = self.find_start_script()
        if path:
            with verbosity(2):
                vprint("Copying start script:", path)
            copy2(path, script_dir)
            return
        with verbosity(2):
            warning(
                "Neither start script or start-command found.\n"
                "Writing a debug start script showing container's environment."
            )
            return self._write_start_script(["env"])

    def _write_start_script(self, start_cmd: list):
        data = deepcopy(self.config.metadata_rendered)
        # data_str = json.dumps(data, indent=4)
        cwd = repr(str(self.source))
        cmd = f"""\
import os

from nua.lib.exec import exec_as_nua
from nua.agent.templates import render_config_templates

render_config_templates({pformat(data)})
exec_as_nua({start_cmd},
            cwd={cwd},
            env=os.environ,)
"""
        script_path = Path(NUA_SCRIPTS_PATH) / "start.py"
        script_path.write_text(cmd)
        with verbosity(2):
            vprint(cmd)

        # anticipate future change for start script:
        if self.is_requiring_templates():
            self._dump_template_data()
        else:
            self._write_shell_start_script(start_cmd)

    def is_requiring_templates(self) -> bool:
        templates_folder = Path("/nua/templates")

        if not templates_folder.is_dir():
            return False

        for file in templates_folder.rglob("*"):
            if not file.is_file():
                continue
            if file.name.startswith("."):
                continue
            return True

        return False

    def _dump_template_data(self):
        with verbosity(1):
            vprint("debug: this app is requiring templates")

        Path("/nua/metadata/template.json").write_text(
            json.dumps(
                self.config.metadata_rendered,
                ensure_ascii=False,
                indent=4,
                sort_keys=True,
            ),
            encoding="utf8",
        )

    def _write_shell_start_script(self, start_cmd: list):
        """Experimental, write a shell start command for dedicated dockerfile."""
        with verbosity(1):
            vprint("debug: write start.sh script")

        work_dir = repr(str(self.source.resolve()))
        # note: /bin/sh (dash on debian) does not support 'source', so using '.'
        cmd = [
            '# executed via "sh -c" as user nua',
            "[ -d /nua/venv ] && . /nua/venv/bin/activate",
            f"cd {work_dir}",
        ]
        cmd.extend(start_cmd)
        script_path = Path(NUA_SCRIPTS_PATH) / "start.sh"
        script_path.write_text("\n".join(cmd), encoding="utf8")
        chown_r(script_path, "nua")
        chmod_r(script_path, 0o755)  # noqa S103

    def find_start_script(self) -> Path | None:
        name = hyphen_get(self.config.build, "start-script")
        if not name:
            name = "start.py"
        path = self.build_dir / "nua" / name
        path = path.absolute().resolve()
        if path.is_file():
            with verbosity(3):
                vprint("found start_script path:", path)
            return path
        return None

    def find_build_script(self) -> Path | None:
        name = hyphen_get(self.config.build, "build-script")
        if not name:
            name = "build.py"
        path = self.build_dir / "nua" / name
        path = path.absolute().resolve()
        if path.is_file():
            with verbosity(3):
                info(f"found build script: {path}")
            return path
        return None

    def run_build_script(self, code_installed: bool) -> bool:
        """Process the 'build.py' script if exists or the 'build' command.

        The script is run from the directory of the nua-config.toml file.
        """
        pip_installed = install_pip_packages(self.config.pip_install)

        if self.config.build_command:
            return self.build_with_command()

        if script_path := self.find_build_script():
            return self.build_with_script(script_path)

        return self.build_with_auto_detection(code_installed, pip_installed)

    def build_with_command(self) -> bool:
        """Process the 'build' commands.

        The script is run from the source directory.
        """
        with chdir(self.source):
            with verbosity(2):
                show("Execution of 'build' command")
            env = dict(os.environ)
            if self.config.build.get("build-as-root", False):
                exec_as_root(self.config.build_command, env=env, timeout=1800)
            else:
                exec_as_nua(self.config.build_command, env=env, timeout=1800)

        return True

    def build_with_script(self, script_path: Path) -> bool:
        """Build with a python script."""
        env = dict(os.environ)
        cmd = f"python {script_path}"
        if self.config.build.get("build-as-root", False):
            exec_as_root(cmd, env=env, timeout=1800)
        else:
            exec_as_nua(cmd, env=env, timeout=1800)

        return True

    def build_with_auto_detection(
        self,
        code_installed: bool,
        pip_installed: bool,
    ) -> bool:
        """Build with a auto detect/install.

        Rem: detect_and_install() is launched as root (the current user),
             it's the responsability of auto installer to switch to user nua.
        """
        if code_installed:
            return detect_and_install(self.source)

        if not any((pip_installed, code_installed)):
            # no package installed through install_pip_packages and
            # no other way. Let's assume there is a local project.
            show("Try install from some local project")
            return detect_and_install(".")

        return False

    def install_project_code(self) -> bool:
        if self.config.src_url:
            self.source = install_source(
                self.config.src_url,
                "/nua/build",
                self.config.app_id,
                self.config.src_checksum,
            )
            return True

        if self.config.project:
            self.source = install_source(
                self.config.project,
                "/nua/build",
                self.config.app_id,
            )
            return True

        if self.config.git_url:
            self.source = install_git_source(
                self.config.git_url, self.config.git_branch, "/nua/build"
            )
            return True

        return False

    def merge_files(self):
        """Copy content of various /nua/build/nua subfolders in /nua."""
        root = Path("/nua/build/nua")
        if not root.is_dir():
            with verbosity(3):
                print("not a directory: /nua/build/nua")
            return

        for item in root.iterdir():
            if item.name == "nua" or not item.is_dir():
                continue
            for file in item.glob("**/*"):
                if not file.is_file():
                    continue
                with verbosity(3):
                    print("merge file:", item)
                target = Path("/nua").joinpath(file.relative_to(root))
                target.parent.mkdir(parents=True, exist_ok=True)
                copy2(file, target.parent)

    def test_build(self):
        """Execute a configured shell command to check build is successful."""
        default = "test -f /nua/metadata/nua-config.*"
        command = self.config.build.get("test", default)
        if not command:
            return

        show("Execution of build test command:", command)
        sh(command, show_cmd=False)


def main() -> None:
    """Setup app in Nua container."""
    if "nua_verbosity" in os.environ:
        set_verbosity(int(os.environ["nua_verbosity"]))
        with verbosity(3):
            info("verbosity:", verbosity_level())

    builder = AppBuilder()
    builder.build()



"""Auto install Python project."""
from pathlib import Path

from nua.lib.actions import build_python

from ..auto_install import register_detector
from .base_detector import BaseDetector


class PythonSource(BaseDetector):
    message: str = "Python source project"
    priority: int = 100

    @classmethod
    def detect(cls) -> bool:
        root = Path(".").resolve()
        python_project_files = ("requirements.txt", "setup.py", "pyproject.toml")
        return any((root / f).exists() for f in python_project_files)

    @classmethod
    def install(cls) -> None:
        build_python(Path("."), "nua")


register_detector(PythonSource)



"""Auto install Nodejs/Yarn."""
from pathlib import Path

from nua.lib.shell import sh

from ..auto_install import register_detector
from .base_detector import BaseDetector


class NodejsYarn(BaseDetector):
    message: str = "Nodejs yarn"
    priority: int = 100
    build_packages: list[str] = [
        "build-essential",
        "python3-dev",
        "libsqlite3-dev",
        "netcat",
        "libicu-dev",
        "libssl-dev",
        "git",
    ]

    @classmethod
    def detect(cls) -> bool:
        root = Path(".").resolve()
        return (root / "yarn.lock").exists()

    @classmethod
    def install(cls) -> None:
        # We install nodejs globally, so if install as user nua we got
        # this error:
        # EACCES: permission denied, open '/root/.config/yarn'
        # fixme: fix global installer path for npm
        #
        # sh("sudo -nu nua yarn install")
        sh("yarn install")


register_detector(NodejsYarn)
